# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0010_merge'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='process_slug',
            field=models.CharField(max_length=200, choices=[('website_enhancement_experiment', 'Website Enhancement Experiment'), (
                'website_enhancement', 'Website Enhancement'), ('doctors', 'Doctors Process')]),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[('website_enhancement_experiment', 'Website Enhancement'), (
                'website_enhancement', 'Website Enhancement'), ('export', 'Export'), ('design', 'Design'), ('content_extraction', ' Content Extraction')]),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='status',
            field=models.IntegerField(
                choices=[(0, 'Processing'), (1, 'Submitted')]),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='worker',
            field=models.ForeignKey(
                on_delete=models.CASCADE, blank=True, null=True, to='orchestra.Worker'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-10-25 08:37
from __future__ import unicode_literals

from django.db import migrations


def copy_staffing_priority(apps, schema_editor):
    Worker = apps.get_model('orchestra', 'Worker')
    for worker in Worker.objects.all():
        # Set worker certification staffing priority
        # the same as original worker staffing priority.
        for certification in worker.certifications.all():
            certification.staffing_priority = worker.staffing_priority
            certification.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0071_workercertification_staffing_priority'),
    ]

    operations = [
        migrations.RunPython(copy_staffing_priority)  # manually-reviewed
    ]



from orchestra.json_schemas.validation import validate_json


class JSONSchemaValidationMixin(object):
    json_schemas = {}

    def save(self, *args, **kwargs):
        for blob_key, json_schema in self.json_schemas.items():
            validate_json(blob_key, json_schema,
                          getattr(self, blob_key, None))
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-08-06 15:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0078_todoqa'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='status',
            field=models.IntegerField(choices=[(0, 'Active'), (2, 'Aborted'), (3, 'Paused')], default=0),
        ),
    ]



import json
import os

from orchestra.core.errors import WorkflowError
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.workflow.directory import parse_workflow_directory

NONEXISTENT_DIR = '/foo/bar/baz/'
TEST_WORKFLOWS_DIR = os.path.abspath(os.path.dirname(__file__))
INVALID_WORKFLOW_DIR = os.path.join(TEST_WORKFLOWS_DIR, 'invalid_dir')
EMPTY_WORKFLOW_DIR = os.path.join(TEST_WORKFLOWS_DIR, 'no_version_dir')
VALID_WORKFLOW_DIR = os.path.join(TEST_WORKFLOWS_DIR, 'test_dir')


class WorkflowDirectoryTestCase(OrchestraTestCase):

    def test_parse_workflow_directory(self):
        """ Ensure that workflow version directories are parsed correctly. """

        # Shouldn't be able to parse a non-existent directory.
        directory_error = 'Workflow directory does not exist.'
        with self.assertRaisesMessage(WorkflowError, directory_error):
            parse_workflow_directory(NONEXISTENT_DIR)

        # Shouldn't be able to parse a directory with no workflow.json.
        with self.assertRaisesMessage(
                WorkflowError,
                'No "workflow.json" manifest file found'):
            parse_workflow_directory(INVALID_WORKFLOW_DIR)

        # Shouldn't be able to parse a directory with no valid versions.
        with self.assertRaisesMessage(
                WorkflowError,
                'Workflow directory {} does not contain any versions'.format(
                    EMPTY_WORKFLOW_DIR)):
            parse_workflow_directory(EMPTY_WORKFLOW_DIR)

        # Should be able to parse a valid directory with two versions.
        parsed = parse_workflow_directory(VALID_WORKFLOW_DIR)
        workflow_json_path = os.path.join(VALID_WORKFLOW_DIR, 'workflow.json')
        with open(workflow_json_path, 'r') as workflow_json_file:
            workflow_data = json.load(workflow_json_file)
        self.assertEqual(workflow_data, parsed['workflow'])

        v1_json_path = os.path.join(VALID_WORKFLOW_DIR, 'v1/version.json')
        with open(v1_json_path, 'r') as v1_json_file:
            v1_data = json.load(v1_json_file)
        self.assertIn(v1_data, parsed['versions'])

        v2_json_path = os.path.join(VALID_WORKFLOW_DIR, 'v2/version.json')
        with open(v2_json_path, 'r') as v2_json_file:
            v2_data = json.load(v2_json_file)
        self.assertIn(v2_data, parsed['versions'])



"""
Validation function for json schemas
"""
from jsonschema import Draft4Validator
from jsonschema import validators


def extend_with_default(validator_class):
    """
    Extends json schema validator so that it fills in default values.

    NOTE(aditya): Copied code from
    https://github.com/b12io/crowdsurfing/blob/master/product/common/json_schema.py
    """
    validate_properties = validator_class.VALIDATORS['properties']

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if 'default' in subschema:
                instance.setdefault(property, subschema['default'])

        for error in validate_properties(
                validator, properties, instance, schema):
            yield error

    return validators.extend(validator_class, {'properties': set_defaults})


DefaultValidatingDraft4Validator = extend_with_default(Draft4Validator)



from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import user_passes_test
from jsonview.decorators import json_view


def is_project_admin(user):
    return (user.groups.filter(name='project_admins').exists() or
            user.is_superuser)


def project_management_api_view_base(func):
    @json_view
    @login_required
    def func_wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return func_wrapper


def project_management_api_view(func):
    @project_management_api_view_base
    @user_passes_test(is_project_admin)
    def func_wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return func_wrapper



from unittest.mock import MagicMock

from apiclient import errors


class FakeImageResponse(object):

    def __init__(self, url):
        self.url = url

        if url is None:
            self.status_code = 400
        else:
            self.status_code = 200

            if 'nocontenttype' not in url:
                self.headers = {
                    'content-type': 'image/jpg'
                }
            else:
                self.headers = {}

    def iter_content(self, size):
        return [b'chunk1']


def _execute_permission_fail():
    return {'id': 'error'}


def _execute_no_error():
    return {'id': 1, 'alternateLink': 'http://a.google.com/link'}


def _execute_error():
    raise errors.HttpError(resp='', content=b'')


def _insert_delete_copy(**kwargs):
    insert_obj = MagicMock()
    body = kwargs.get('body')
    if ((body and body.get('title') and 'error' in body['title']) or
            kwargs.get('fileId') == 'error'):
        insert_obj.execute = _execute_error
    elif (kwargs.get('body') and
          kwargs['body'].get('title') and
          'permission_fail' in kwargs['body']['title']):
        insert_obj.execute = _execute_permission_fail
    else:
        insert_obj.execute = _execute_no_error
    return insert_obj


def _files():
    files_obj = MagicMock()
    files_obj.insert = _insert_delete_copy
    files_obj.delete = _insert_delete_copy
    files_obj.copy = _insert_delete_copy
    return files_obj


_drive_service = MagicMock()
_drive_service.files = _files
_drive_service.permissions = _files
mock_create_drive_service = MagicMock(return_value=_drive_service)


def fake_image_get(url, *args, **kwargs):
    return FakeImageResponse(url)



from unittest.mock import patch

from orchestra.core.errors import MachineExecutionError
from orchestra.machine_tasks import execute
from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.task_lifecycle import create_subsequent_tasks


class MachineTaskTestCase(OrchestraTestCase):
    def setUp(self):
        super().setUp()
        setup_models(self)

        machine_workflow_version = (
            self.workflow_versions['machine_workflow_version'])
        self.step = machine_workflow_version.steps.get(slug='machine_step')
        self.assertIsNotNone(self.step)
        self.step.execution_function = {
            'path': 'orchestra.tests.helpers.workflow.machine_task_function',
        }
        self.step.save()
        patcher = patch(
            'orchestra.tests.helpers.workflow.machine_task_function',
            return_value={'data': ''})
        self.machine_function_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.project = ProjectFactory(
            workflow_version=machine_workflow_version)
        create_subsequent_tasks(self.project)
        self.task = self.project.tasks.first()
        self.assertIsNotNone(self.task)

        # Since create_subsequent_tasks will automatically run the machine task
        # we need to reset the task each time it's called
        self._reset_task()

    def _reset_task(self):
        self.task.status = Task.Status.AWAITING_PROCESSING
        self.task.save()
        self.task.assignments.all().delete()
        self.machine_function_mock.call_count = 0

    def _assert_correct_machine_task_state(
            self, assignment_status, task_status):
        self.task.refresh_from_db()
        assignment = self.task.assignments.first()
        self.assertEqual(assignment.status, assignment_status)
        self.assertEqual(self.task.status, task_status)

        # Check correct iteration state
        self.assertEqual(assignment.iterations.count(), 1)
        iteration = assignment.iterations.first()
        if task_status == Task.Status.COMPLETE:
            expected_status = Iteration.Status.REQUESTED_REVIEW
            expected_data = assignment.in_progress_task_data
            self.assertIsNotNone(iteration.end_datetime)
        else:
            expected_status = Iteration.Status.PROCESSING
            expected_data = {}
            self.assertIsNone(iteration.end_datetime)

        self.assertEqual(iteration.status, expected_status)
        self.assertEqual(iteration.submitted_data, expected_data)

        # Assert that machine function is called once and there is only one
        # assignment regardless of state
        self.assertEqual(self.machine_function_mock.call_count, 1)
        self.assertEqual(self.task.assignments.count(), 1)

    def test_new_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    def test_already_processing_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)

        # Pretend that the task is still processing
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        assignment = self.task.assignments.first()
        assignment.status = TaskAssignment.Status.PROCESSING
        assignment.save()

        # Reset iteration for assignment
        assignment.iterations.all().delete()
        Iteration.objects.create(
            assignment=assignment,
            start_datetime=assignment.start_datetime)

        # Another machine attempts to perform task
        with self.assertRaises(MachineExecutionError):
            execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.PROCESSING, Task.Status.PROCESSING)

    def test_already_completed_task(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)
        with self.assertRaises(MachineExecutionError):
            execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    @patch('orchestra.machine_tasks.logger')
    def test_marking_failed_task_assignment(self, mock_logger):
        self.machine_function_mock.side_effect = Exception('Function failed.')
        self._reset_task()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.FAILED, Task.Status.PROCESSING)
        mock_logger.exception.assert_called_with('Machine task has failed')
        self.machine_function_mock.side_effect = None

    def test_reassigning_failed_task_assignment(self):
        self._reset_task()
        execute(self.project.id, self.step.slug)

        # Pretend that the machine task failed
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        assignment = self.task.assignments.first()
        assignment.status = TaskAssignment.Status.FAILED
        assignment.save()
        self.assertEqual(self.machine_function_mock.call_count, 1)
        self.machine_function_mock.call_count = 0

        # New machine picks up the failed task
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.COMPLETE)

    def test_aborted_project(self):
        self._reset_task()
        self.project.status = Project.Status.ABORTED
        self.project.save()
        self.task.status = Task.Status.PROCESSING
        self.task.save()
        execute(self.project.id, self.step.slug)
        self._assert_correct_machine_task_state(
            TaskAssignment.Status.SUBMITTED, Task.Status.ABORTED)



import json

from django.test import override_settings

from orchestra.models import Step
from orchestra.models import Task
from orchestra.models import Project
from orchestra.models import TaskAssignment
from orchestra.tests.helpers import OrchestraTransactionTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.iterations import verify_iterations
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.task_lifecycle import create_subsequent_tasks


class DashboardTestCase(OrchestraTransactionTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)

    @override_settings(MACHINE_STEP_SCHEDULER={
        'path': ('orchestra.utils.machine_step_scheduler.'
                 'SynchronousMachineStepScheduler'),
    })
    def test_task_creation(self):
        """
        Test human and machine task creation
        """
        Task.objects.filter(status=Task.Status.AWAITING_PROCESSING).delete()

        project = self.projects['test_human_and_machine']
        self.assertEqual(Task.objects.filter(project=project).count(),
                         0)
        create_subsequent_tasks(project)

        # Human Task was created
        self.assertEqual(Task.objects.filter(project=project).count(),
                         1)

        response = (self.clients[0].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))
        self.assertEqual(response.status_code, 200)

        human_step = self.workflow_steps['test_workflow_2']['step4']
        task = Task.objects.get(step=human_step, project=project)
        data = {'submit_key1': 'submit_val1'}

        # user 0 submits a task
        response = self._submit_assignment(self.clients[0], task.id, data=data)
        self.assertEqual(response.status_code, 200)

        # Machine Task was created
        self.assertEqual(Task.objects.filter(project=project).count(),
                         2)
        machine_step = self.workflow_steps['test_workflow_2']['simple_machine']
        machine_task_assignment = (
            TaskAssignment.objects
            .filter(task__step=machine_step,
                    task__project=project)[0])

        self.assertTrue(machine_task_assignment.status,
                        TaskAssignment.Status.SUBMITTED)

        self.assertTrue(machine_task_assignment.in_progress_task_data,
                        {'simple': 'json'})

        self.assertTrue(machine_task_assignment.task.status,
                        Task.Status.COMPLETE)

    def test_index(self):
        response = self.clients[0].get('/orchestra/app/')
        self.assertEqual(response.status_code, 200)

    def test_status(self):
        response = self.clients[0].get('/orchestra/status/')
        self.assertEqual(response.status_code, 200)

    def test_dashboard_tasks(self):
        self._check_client_dashboard_state(self.clients[0], 'pending_review')

    def test_entry_level_task_assignment(self):
        response = (self.clients[2].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))

        # no task available for user3 because no entry-level certification
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'No worker certificates')

        # user 0 only has certification for entry level tasks
        response = (self.clients[0].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        task = Task.objects.get(id=returned['id'])
        self.assertEqual({
            'id': task.id,
            'assignment_id': task.assignments.get(worker=self.workers[0]).id,
            'step': task.step.slug,
            'project': task.project.workflow_version.slug,
            'detail': task.project.short_description
        }, returned)

        # task assignment for invalid id should give bad request
        self._verify_bad_task_assignment_information(
            self.clients[0], {'task_id': -1},
            'Task matching query does not exist.')

        # task assignment for user3 not assigned to a task
        self._verify_bad_task_assignment_information(
            self.clients[2], {'task_id': task.id},
            'Worker is not associated with task')

        # task assignment is assigned to user 0
        self._verify_good_task_assignment_information(
            self.clients[0], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Processing', False,
            False, {}, self.workers[0])

        # no more tasks for user 0 left
        response = (self.clients[0].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'No task')

    def _check_client_dashboard_state(self, client, non_empty_status):
        response = client.get('/orchestra/api/interface/dashboard_tasks/')
        returned = load_encoded_json(response.content)
        non_empty_tasks = [task for task in returned['tasks']
                           if task['state'] == non_empty_status]
        empty_tasks = [task for task in returned['tasks']
                       if task['state'] != non_empty_status]
        self.assertGreater(len(non_empty_tasks), 0)
        self.assertEqual(len(empty_tasks), 0)

    def test_reviewer_task_assignment(self):
        # there is a review task for user 1
        response = (self.clients[1].get(
            '/orchestra/api/interface/new_task_assignment/reviewer/'))
        self.assertEqual(response.status_code, 200)

        returned = load_encoded_json(response.content)
        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': returned['id']},
            Task.objects.get(id=returned['id']).project.short_description,
            'Processing', 'Reviewing', True,
            False, {'test_key': 'test_value'}, self.workers[1])

    def test_save_entry_level_task_assignment(self):
        # user 0 only has certification for entry level tasks
        response = (self.clients[0].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)
        task = Task.objects.get(id=returned['id'])

        # incorrect task id
        response = self.clients[0].post(
            '/orchestra/api/interface/save_task_assignment/',
            json.dumps({'task_id': -1,
                        'task_data': 'test'}),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'No task for given id')

        # user does not have a permission to save
        response = self.clients[1].post(
            '/orchestra/api/interface/save_task_assignment/',
            json.dumps({'task_id': task.id,
                        'task_data': 'test'}),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker is not associated with task')

        # save new info
        new_data = {'new_test_key': 'new_test_value'}
        response = self.clients[0].post(
            '/orchestra/api/interface/save_task_assignment/',
            json.dumps({'task_id': task.id,
                        'task_data': new_data}),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)

        self._verify_good_task_assignment_information(
            self.clients[0], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Processing', False,
            False, new_data, self.workers[0])

    def test_save_reviewer_task_assignment(self):
        new_data = {'new_test_key': 'new_test_value'}

        # get a reviewer task
        response = (self.clients[1].get(
            '/orchestra/api/interface/new_task_assignment/reviewer/'))
        self.assertEqual(response.status_code, 200)

        returned = load_encoded_json(response.content)

        task = Task.objects.get(id=returned['id'])

        # entry level worker can't update the data
        response = self.clients[0].post(
            '/orchestra/api/interface/save_task_assignment/',
            json.dumps({'task_id': task.id,
                        'task_data': new_data}),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker is not allowed to save')

        reviewer_data = {'reviewer_key': 'reviewer_value'}

        # reviewer can update the data
        response = self.clients[1].post(
            '/orchestra/api/interface/save_task_assignment/',
            json.dumps({'task_id': task.id,
                        'task_data': reviewer_data}),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)

        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Reviewing', True,
            False, reviewer_data, self.workers[1])

    def test_submit_entry_level_task_assignment(self):
        # user 0 only has certification for entry level tasks
        response = (self.clients[0].get(
            '/orchestra/api/interface/new_task_assignment/entry_level/'))
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)
        task = Task.objects.get(id=returned['id'])

        verify_iterations(task.id)

        # user is not assigned to a task
        response = self._submit_assignment(self.clients[1], task.id)
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Task assignment with worker is in broken state.')

        # task does not exist
        response = self._submit_assignment(self.clients[1], -1)
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'No task for given id')

        # user 0 can only submit a task not reject
        response = self._submit_assignment(
            self.clients[0], task.id, command='reject')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Task not in rejectable state.')

        # user 0 can't call illegal commands
        response = self._submit_assignment(
            self.clients[0], task.id, command='approve')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Illegal command')
        data = {'submit_key1': 'submit_val1'}

        # user 0 can't submit a task if its submission prerequisites aren't
        # complete
        step = task.step
        step.submission_depends_on = [
            Step.objects.create(
                workflow_version=step.workflow_version,
                slug='imaginary_test_step',
                is_human=True,
            )
        ]
        step.save()
        response = self._submit_assignment(
            self.clients[0], task.id)
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Submission prerequisites are not complete.')
        step.submission_depends_on.set([])
        step.save()

        data = {'submit_key1': 'submit_val1'}
        # user 0 submits a task
        response = self._submit_assignment(
            self.clients[0], task.id, data=data)
        self.assertEqual(response.status_code, 200)

        self._verify_good_task_assignment_information(
            self.clients[0], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Pending Review', False,
            True, data, self.workers[0])

        # Check that iteration has correct submitted state
        verify_iterations(task.id)

        # user cannot resubmit a task
        response = self._submit_assignment(
            self.clients[0], task.id)
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker is not allowed to submit')

    def test_submit_reviewer_task_assignment(self):
        data = {'submit_key1': 'submit_val1'}

        # user 1 is picking up a task as a reviewer
        response = (self.clients[1].get(
            '/orchestra/api/interface/new_task_assignment/reviewer/'))
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)
        task_id = returned['id']
        task = Task.objects.get(id=returned['id'])
        self.assertEqual(task.assignments.count(), 2)
        self.assertEqual(task_id, self.tasks['review_task'].id)

        verify_iterations(task.id)

        rejected_data = {'rejected_key': 'rejected_val'}

        # user 0 can retrieve data, but should see a read-only interface
        self._verify_good_task_assignment_information(
            self.clients[0], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Reviewing', False,
            True, {'test_key': 'test_value'}, self.workers[0])

        # user 1 should be able to review the post
        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Reviewing', True,
            False, {'test_key': 'test_value'}, self.workers[1])

        # user 1 rejects a task
        response = self._submit_assignment(
            self.clients[1], task_id, data=rejected_data, command='reject')
        self.assertEqual(response.status_code, 200)

        verify_iterations(task.id)

        # user 0 should have the task back
        self._verify_good_task_assignment_information(
            self.clients[0], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Post-review Processing', False,
            False, rejected_data, self.workers[0])

        # user 1 should no longer be able to modify the post
        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Post-review Processing', True,
            True, rejected_data, self.workers[1])

        # user 0 submits an updated data
        response = self._submit_assignment(
            self.clients[0], task_id, data=data)
        self.assertEqual(response.status_code, 200)

        verify_iterations(task.id)

        # check if the data is saved
        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': task.id},
            task.project.short_description,
            'Processing', 'Reviewing', True,
            False, data, self.workers[1])

        accepted_data = {'accepted_key': 'accepted_val'}
        # user 1 accepts a task
        response = self._submit_assignment(
            self.clients[1], task_id, data=accepted_data, command='accept')
        self.assertEqual(response.status_code, 200)

        verify_iterations(task.id)

        # check if the accepted_data is saved
        # and task is pending for a second review.
        self._verify_good_task_assignment_information(
            self.clients[1], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Pending Review', True,
            True, accepted_data, self.workers[1])

        # make sure a task can't be submitted twice
        response = self._submit_assignment(
            self.clients[1], task_id, command='accept')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Worker is not allowed to submit')

        # user 3 is picking up a task as a reviewer
        response = (self.clients[3].get(
            '/orchestra/api/interface/new_task_assignment/reviewer/'))
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['id'],
                         task.id)

        verify_iterations(task.id)

        rejected_data = {'rejected_key': 'rejected_val'}

        # user 3 rejects a task
        response = self._submit_assignment(
            self.clients[3], task_id, data=rejected_data, command='reject')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        verify_iterations(task.id)

        # check if the rejected_data is saved
        self._verify_good_task_assignment_information(
            self.clients[3], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Post-review Processing', True,
            True, rejected_data, self.workers[3])

        # check client dashboards
        self._check_client_dashboard_state(self.clients[0], 'pending_review')

        self._check_client_dashboard_state(self.clients[1], 'returned')

        self._check_client_dashboard_state(self.clients[3],
                                           'pending_processing')

        response = self._submit_assignment(
            self.clients[1], task.id)
        self.assertEqual(response.status_code, 200)

        verify_iterations(task.id)

        # check if client dashboards were updated
        self._check_client_dashboard_state(self.clients[0], 'pending_review')

        self._check_client_dashboard_state(self.clients[1], 'pending_review')

        self._check_client_dashboard_state(self.clients[3], 'in_progress')

        # check if the accepted_data is saved
        response = self._submit_assignment(
            self.clients[3], task_id, data=accepted_data, command='accept')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        verify_iterations(task.id)

        # check if task is complete
        self._verify_good_task_assignment_information(
            self.clients[3], {'task_id': task.id},
            task.project.short_description,
            'Submitted', 'Complete', True,
            True, accepted_data, self.workers[3])

        # check that reviewer cannot reaccept task
        response = self._submit_assignment(
            self.clients[3], task_id, data=accepted_data, command='accept')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         'Task already completed')

    def _verify_bad_task_assignment_information(
            self, client, post_data, error_message):
        response = client.post(
            '/orchestra/api/interface/task_assignment_information/',
            json.dumps(post_data),
            content_type='application/json')
        self.assertEqual(response.status_code, 400)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned['message'],
                         error_message)

    def _verify_good_task_assignment_information(
            self, client, post_data, project_description,
            assignment_status, task_status, is_reviewer,
            is_read_only, task_data, worker):
        response = client.post(
            '/orchestra/api/interface/task_assignment_information/',
            json.dumps(post_data),
            content_type='application/json')
        self.assertEqual(response.status_code, 200)

        task = Task.objects.get(id=post_data['task_id'])
        returned = load_encoded_json(response.content)
        expected = {
            'project': {'details': project_description,
                        'id': task.project.id,
                        'project_data': {},
                        'status': dict(
                            Project.STATUS_CHOICES)[task.project.status],
                        'team_messages_url': None},
            'status': assignment_status,
            'task': {'data': task_data, 'status': task_status},
            'task_id': task.id,
            'assignment_id': task.assignments.get(worker=worker).id,
            'workflow': {
                'slug': 'w1', 'name': 'Workflow One',
            },
            'workflow_version': {
                'slug': 'test_workflow', 'name': 'The workflow'},
            'prerequisites': {},
            'step': {'slug': 'step1', 'name': 'The first step'},
            'is_reviewer': is_reviewer,
            'is_read_only': is_read_only,
            'is_project_admin': False,
            'worker': {
                'username': worker.user.username,
                'first_name': worker.user.first_name,
                'last_name': worker.user.last_name,
            }
        }

        self.assertEqual(returned, expected)



import logging

from django.db import transaction

from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse

logger = logging.getLogger(__name__)


@transaction.atomic
def mark_worker_as_winner(worker, task, required_role_counter,
                          staffing_request_inquiry):
    staffbot_request = (
        StaffBotRequest.objects
        .filter(task=task, required_role_counter=required_role_counter)
        .order_by('-created_at'))

    # Check whether staffbot request was sent out for this task
    if not staffbot_request.exists():
        return

    if staffing_request_inquiry:
        staffbot_request = staffing_request_inquiry.request
    else:
        staffbot_request = staffbot_request.first()

    staffbot_request.status = StaffBotRequest.Status.COMPLETE.value
    staffbot_request.save()

    # Mark everyone else as non-winner
    StaffingResponse.objects.filter(
        request_inquiry__request=staffbot_request).update(is_winner=False)

    # If staffing request inquiry provided
    if staffing_request_inquiry:
        staffing_response = staffing_request_inquiry.responses.all()
    else:
        staffing_response = StaffingResponse.objects.filter(
            request_inquiry__request=staffbot_request,
            request_inquiry__communication_preference__worker=worker)

    if staffing_response.exists():
        staffing_response = staffing_response.first()
        staffing_response.is_available = True
        staffing_response.is_winner = True
        staffing_response.save()
    else:
        # This branch occurs if a worker was assigned to a task that has
        # an open StaffBot Request, but StaffBot hasn't reached out
        # to the Worker yet. We fib a little and create in inquiry/response
        # for the worker so they can be marked as a winner.
        comm_pref = CommunicationPreference.objects.filter(
            worker=worker)
        if not comm_pref.exists():
            logger.error('Worker {} does not have a communication '
                         'preferences setup'.format(worker))
            return

        comm_pref = comm_pref.first()
        comm_method = StaffingRequestInquiry.CommunicationMethod.SLACK.value

        if not staffing_request_inquiry:
            staffing_request_inquiry = (
                StaffingRequestInquiry.objects
                .create(request=staffbot_request,
                        communication_preference=comm_pref,
                        communication_method=comm_method))

        StaffingResponse.objects.create(
            request_inquiry=staffing_request_inquiry,
            is_available=True,
            is_winner=True)



import logging

from rest_framework import generics
from rest_framework import permissions
from rest_framework.response import Response
from jsonview.exceptions import BadRequest

from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.models import TodoListTemplate
from orchestra.models import Worker
from orchestra.todos.serializers import TodoSerializer
from orchestra.todos.serializers import TodoWithQASerializer
from orchestra.todos.serializers import TodoQASerializer
from orchestra.todos.serializers import TodoListTemplateSerializer
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.todos.api import add_todolist_template
from orchestra.utils.decorators import api_endpoint
from orchestra.todos.auth import IsAssociatedWithTodosProject
from orchestra.todos.auth import IsAssociatedWithProject
from orchestra.todos.auth import IsAssociatedWithTask

logger = logging.getLogger(__name__)


@api_endpoint(methods=['POST'],
              permissions=(IsAssociatedWithProject,),
              logger=logger)
def update_todos_from_todolist_template(request):
    todolist_template_slug = request.data.get('todolist_template')
    task_id = request.data.get('task')
    try:
        add_todolist_template(todolist_template_slug, task_id)
        project = Task.objects.get(id=task_id).project
        todos = Todo.objects.filter(
            task__project__id=int(project.id)).order_by('-created_at')
        serializer = TodoSerializer(todos, many=True)
        return Response(serializer.data)
    except TodoListTemplate.DoesNotExist:
        raise BadRequest('TodoList Template not found for the given slug.')


@api_endpoint(methods=['GET'],
              permissions=(permissions.IsAuthenticated,
                           IsAssociatedWithTask),
              logger=logger)
def worker_task_recent_todo_qas(request):
    """
    Returns TodoQA recommendations for the requesting user and task.

    The TodoQAs for the recommendation are selected using the following logic:
    1. If the given task has TodoQAs, use them.
    2. Otherwise, use the TodoQAs from the requesting user's most recent
    task with matching task slug.
    """
    task_id = request.query_params.get('task')
    task_todo_qas = TodoQA.objects.filter(todo__task=task_id)

    if task_todo_qas.exists():
        todo_qas = task_todo_qas
    else:
        task = Task.objects.get(pk=task_id)
        most_recent_worker_task_todo_qa = TodoQA.objects.filter(
            todo__task__assignments__worker__user=request.user,
            todo__task__step__slug=task.step.slug
        ).order_by('-created_at').first()

        if most_recent_worker_task_todo_qa:
            todo_qas = TodoQA.objects.filter(
                todo__task=most_recent_worker_task_todo_qa.todo.task,
                approved=False)
        else:
            todo_qas = TodoQA.objects.none()

    todos_recommendation = {todo_qa.todo.description: TodoQASerializer(
        todo_qa).data for todo_qa in todo_qas}
    return Response(todos_recommendation)


class TodoList(generics.ListCreateAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithProject)

    queryset = Todo.objects.all()

    def get_serializer_class(self):
        # Only include todo QA data for users in the `project_admins` group.
        if self.request.user.groups.filter(
                name='project_admins').exists():
            return TodoWithQASerializer
        else:
            return TodoSerializer

    def get_queryset(self):
        queryset = Todo.objects.all()
        project_id = self.request.query_params.get('project', None)
        if project_id is not None:
            queryset = queryset.filter(task__project__id=int(project_id))
        queryset = queryset.order_by('-created_at')
        return queryset

    def perform_create(self, serializer):
        todo = serializer.save()
        sender = Worker.objects.get(
            user=self.request.user).formatted_slack_username()
        recipients = ' & '.join(
            assignment.worker.formatted_slack_username()
            for assignment in todo.task.assignments.all()
            if assignment and assignment.worker)
        message = '{} has created a new todo `{}` for {}.'.format(
            sender,
            todo.description,
            recipients if recipients else '`{}`'.format(todo.task.step.slug))
        message_experts_slack_group(
            todo.task.project.slack_group_id, message)


class TodoDetail(generics.RetrieveUpdateDestroyAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithTodosProject)
    queryset = Todo.objects.all()
    serializer_class = TodoSerializer

    def perform_update(self, serializer):
        old_todo = self.get_object()
        todo = serializer.save()
        sender = Worker.objects.get(
            user=self.request.user).formatted_slack_username()

        if old_todo.completed != todo.completed:
            todo_change = 'complete' if todo.completed else 'incomplete'
        elif old_todo.skipped_datetime != todo.skipped_datetime:
            todo_change = 'not relevant' \
                if todo.skipped_datetime else 'relevant'
        else:
            # When activity_log is updated, `todo_change = None`
            # to avoid triggering any slack messages
            todo_change = None

        # To avoid Slack noise, only send updates for changed TODOs with
        # depth 0 (no parent) or 1 (no grantparent).
        if todo_change and \
                (not (todo.parent_todo and todo.parent_todo.parent_todo)):
            message = '{} has marked `{}` as `{}`.'.format(
                sender,
                todo.description,
                todo_change)
            message_experts_slack_group(
                todo.task.project.slack_group_id, message)


class TodoQADetail(generics.UpdateAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithTodosProject)

    serializer_class = TodoQASerializer
    queryset = TodoQA.objects.all()


class TodoQAList(generics.CreateAPIView):
    permission_classes = (permissions.IsAuthenticated,
                          IsAssociatedWithProject)

    serializer_class = TodoQASerializer
    queryset = TodoQA.objects.all()


class TodoListTemplateDetail(generics.RetrieveUpdateAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    serializer_class = TodoListTemplateSerializer
    queryset = TodoListTemplate.objects.all()


class TodoListTemplateList(generics.ListCreateAPIView):
    permission_classes = (permissions.IsAuthenticated,)

    serializer_class = TodoListTemplateSerializer
    queryset = TodoListTemplate.objects.all()



from django.apps import apps as django_apps
from django.conf import settings
from django.db import transaction

from orchestra.core.errors import WorkflowError
from orchestra.models import Certification
from orchestra.models import Step
from orchestra.models import TodoListTemplate
from orchestra.models import Workflow
from orchestra.models import WorkflowVersion
from orchestra.workflow.defaults import get_default_assignment_policy
from orchestra.workflow.defaults import get_default_creation_policy
from orchestra.workflow.defaults import get_default_review_policy
from orchestra.workflow.directory import parse_workflow_directory


def get_workflow_version_slugs():
    versions = {}
    for app_name in settings.ORCHESTRA_WORKFLOWS:
        # App label is the last part of the app name by default
        app_label = app_name.split('.')[-1]
        workflow_directory = django_apps.get_app_config(app_label).path
        data = parse_workflow_directory(workflow_directory)
        workflow_slug = data['workflow']['slug']
        if versions.get(workflow_slug) is not None:
            raise WorkflowError('Workflow {} present in multiple apps: {}, {}'
                                .format(workflow_slug,
                                        versions[workflow_slug]['app_label'],
                                        app_label))
        else:
            versions[workflow_slug] = {
                'app_label': app_label,
                'versions': (version['slug'] for version in data['versions'])
            }
    return versions


@transaction.atomic
def load_workflow(app_label, version_slug, force=False):
    workflow_directory = django_apps.get_app_config(app_label).path
    data = parse_workflow_directory(workflow_directory)

    # Create the workflow object if it doesn't exist
    workflow_data = data['workflow']
    workflow, workflow_created = Workflow.objects.update_or_create(
        slug=workflow_data['slug'],
        defaults={
            'name': workflow_data['name'],
            'description': workflow_data['description'],
            'code_directory': workflow_directory,
            'sample_data_load_function': workflow_data.get(
                'sample_data_load_function')
        }
    )

    # Create all certifications for the workflow
    for certification_data in workflow_data['certifications']:
        Certification.objects.update_or_create(
            slug=certification_data['slug'],
            workflow=workflow,
            defaults={
                'name': certification_data['name'],
                'description': certification_data['description'],
            }
        )

    # Create the certification dependencies once all certs are in the db
    # Allow updating these over time so that a workflow's certifications can
    # evolve. This means that the user is responsible for ensuring that all
    # workers have the proper certifications after updating these dependencies.
    for certification_data in workflow_data['certifications']:
        certification = Certification.objects.get(
            slug=certification_data['slug'],
            workflow=workflow
        )
        required_certification_slugs = certification_data.get(
            'required_certifications', [])
        required_certifications = Certification.objects.filter(
            workflow=workflow,
            slug__in=required_certification_slugs
        )
        if required_certifications.count() != len(
                required_certification_slugs):
            raise WorkflowError(
                'Certification {} requires non-existent certification.'
                .format(certification_data['slug']))
        certification.required_certifications.set(
            list(required_certifications))

    # Load the desired versions
    desired_versions = [version_data for version_data in data['versions']
                        if version_data['slug'] == version_slug]
    if len(desired_versions) != 1:
        raise WorkflowError('Invalid version requested: {}'
                            .format(version_slug))
    load_workflow_version(desired_versions[0], workflow, force=force)


def load_workflow_version(version_data, workflow, force=False):
    # Create the version object
    version, version_created = WorkflowVersion.objects.update_or_create(
        slug=version_data['slug'],
        workflow=workflow,
        defaults={
            'name': version_data['name'],
            'description': version_data['description'],
            'sanity_checks': version_data.get('sanity_checks', {}),
        }
    )

    if not version_created:
        if not force:
            # It is safe to error out after modifying the DB because
            # all of this code is wrapped in a transaction by load_workflow.
            raise WorkflowError('Version {} already exists'
                                .format(version_data['slug']))

        # Check that the versions are safe to merge
        new_step_slugs = set(step['slug'] for step in version_data['steps'])
        old_step_slugs = set(
            Step.objects
            .filter(workflow_version=version)
            .values_list('slug', flat=True)
        )
        if new_step_slugs != old_step_slugs:
            raise WorkflowError('Even with --force, you cannot change the '
                                'steps of a workflow. Drop and recreate the '
                                'database to reset, or create a new version '
                                'for your workflow.')

    # Create or update the version steps.
    old_creation_dependencies = {}
    old_submission_dependencies = {}
    for step_data in version_data['steps']:
        is_human = step_data.get('is_human', True)
        step, step_created = Step.objects.update_or_create(
            slug=step_data['slug'],
            workflow_version=version,
            defaults={
                'name': step_data['name'],
                'description': step_data['description'],
                'is_human': is_human,
                'detailed_description_function': step_data.get(
                    'detailed_description_function', {}),
                'execution_function': step_data.get('execution_function', {}),
                'review_policy': step_data.get(
                    'review_policy',
                    get_default_review_policy(is_human)),
                'assignment_policy': step_data.get(
                    'assignment_policy',
                    get_default_assignment_policy(is_human)),
                'creation_policy': step_data.get(
                    'creation_policy',
                    get_default_creation_policy()),
                'user_interface': step_data.get('user_interface', {}),
            }
        )
        if not step_created:
            old_creation_dependencies[step_data['slug']] = set(
                step.creation_depends_on.values_list('slug', flat=True))
            old_submission_dependencies[step_data['slug']] = set(
                step.submission_depends_on.values_list('slug', flat=True))

        # Don't prevent updates to these, because we want to allow
        # certifications to evolve over the lifetime of a workflow.
        _set_step_relations(step, step_data, 'required_certifications',
                            Certification, workflow=workflow)

    # Set up step dependencies once the steps objects are in the DB.
    for step_data in version_data['steps']:
        step_slug = step_data['slug']
        step = Step.objects.get(
            slug=step_slug,
            workflow_version=version
        )

        # Set step creation dependencies.
        _verify_dependencies_not_updated(
            step_data,
            'creation_depends_on',
            old_creation_dependencies.get(step_slug)
        )
        _set_step_relations(step, step_data, 'creation_depends_on', Step,
                            workflow_version=version)

        _set_step_relations(step, step_data, 'submission_depends_on', Step,
                            workflow_version=version)

        _set_step_relations(step, step_data, 'todolist_templates_to_apply',
                            TodoListTemplate)


def _verify_dependencies_not_updated(step_data, dependency_attr,
                                     old_dependencies):
    new_dependencies = set(step_data.get(dependency_attr, []))
    if old_dependencies is not None and old_dependencies != new_dependencies:
        raise WorkflowError(
            'Even with --force, you cannot change the topology of a workflow. '
            'Drop and recreate the database to reset, or create a new '
            'version for your workflow.')


def _set_step_relations(step, step_data, relation_attr, relation_model,
                        **model_filters):
    relation_slugs = set(step_data.get(relation_attr, []))
    relations = list(relation_model.objects.filter(
        slug__in=relation_slugs, **model_filters))
    if len(relations) != len(relation_slugs):
        raise WorkflowError(
                            '{}.{} contains a non-existent slug.'
                            .format(step_data['slug'], relation_attr))
    getattr(step, relation_attr).set(relations)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-27 17:43
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0047_auto_20160427_1412'),
        ('orchestra', '0047_worker_phone'),
    ]

    operations = [
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-04 17:31
from __future__ import unicode_literals

import jsonfield.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0058_staffingrequest_required_role'),
    ]

    operations = [
        migrations.AddField(
            model_name='step',
            name='detailed_description_function',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



from collections import Counter
from datetime import timedelta
from unittest.mock import patch

from orchestra.bots.sanitybot import create_and_handle_sanity_checks
from orchestra.models import SanityCheck
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.task_lifecycle import create_subsequent_tasks


class SanityBotTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)

    @patch('orchestra.bots.sanitybot.message_experts_slack_group')
    def test_create_and_handle_sanity_checks(self, mock_slack):
        # Initialize relevant project.
        create_subsequent_tasks(self.projects['sanitybot'])
        # Initialize irrelevant project.
        create_subsequent_tasks(self.projects['test_human_and_machine'])

        # No sanity checks exist, and Slack hasn't received a message.
        sanity_checks = SanityCheck.objects.all()
        self.assertEqual(sanity_checks.count(), 0)
        self.assertEqual(mock_slack.call_count, 0)

        create_and_handle_sanity_checks()
        # Of the four sanity checks for this project, three were
        # triggered by orchestra.tests.helpers.workflow.check_project.
        sanity_checks = SanityCheck.objects.all()
        self.assertEqual(
            dict(Counter(sanity_checks.values_list('check_slug', flat=True))),
            {
                'frequently_repeating_check': 1,
                'infrequently_repeating_check': 1,
                'onetime_check': 1
            })
        for sanity_check in sanity_checks:
            self.assertEqual(
                sanity_check.project.workflow_version.slug,
                'sanitybot_workflow')
        # Three slack messages for three sanity checks.
        self.assertEqual(mock_slack.call_count, 3)

        # Too little time has passed, so we expect no new sanity checks.
        create_and_handle_sanity_checks()
        # Still only three sanity check exists for the relevant project.
        sanity_checks = SanityCheck.objects.all()
        self.assertEqual(sanity_checks.count(), 3)
        # Slack didn't get called again.
        self.assertEqual(mock_slack.call_count, 3)

        # Mark the sanity checks as having happened two days ago. Only
        # the `frequently_repeating_check` should trigger.
        for sanity_check in sanity_checks:
            sanity_check.created_at = (
                sanity_check.created_at - timedelta(days=2))
            sanity_check.save()
        create_and_handle_sanity_checks()
        # Look for the three old sanity checks and a new
        # `frequently_repeating_check`.
        sanity_checks = SanityCheck.objects.all()
        self.assertEqual(
            dict(Counter(sanity_checks.values_list('check_slug', flat=True))),
            {
                'frequently_repeating_check': 2,
                'infrequently_repeating_check': 1,
                'onetime_check': 1
            })
        for sanity_check in sanity_checks:
            self.assertEqual(
                sanity_check.project.workflow_version.slug,
                'sanitybot_workflow')
        # Slack got called another time.
        self.assertEqual(mock_slack.call_count, 4)

        create_and_handle_sanity_checks()
        # Still only four sanity checks exists for the relevant project.
        sanity_checks = SanityCheck.objects.all()
        self.assertEqual(sanity_checks.count(), 4)
        # Slack didn't get called again.
        self.assertEqual(mock_slack.call_count, 4)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-08-14 17:59
from __future__ import unicode_literals

from django.db import migrations, models
from django.db.models import Count

PROJECT_STATUS_ACTIVE = 0
PROJECT_STATUS_COMPLETED = 4

TASK_STATUS_COMPLETED = 5

def mark_completed_projects(apps, schema_editor):
    Project = apps.get_model('orchestra', 'Project')
    Task = apps.get_model('orchestra', 'Task')

    projects = Project.objects.filter(status=PROJECT_STATUS_ACTIVE)
    in_progress_projects = (
        Task.objects.exclude(status=TASK_STATUS_COMPLETED)
        .values_list('project', flat=True))
    completed_projects = (
        projects.annotate(num_tasks=Count('tasks'))
        .filter(num_tasks__gt=0)
        .exclude(id__in=in_progress_projects))

    for project in completed_projects:
        if (project.status == PROJECT_STATUS_ACTIVE):
            project.status = PROJECT_STATUS_COMPLETED
            project.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0079_auto_20180806_1528'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='status',
            field=models.IntegerField(choices=[(0, 'Active'), (2, 'Aborted'), (3, 'Paused'), (4, 'Completed')], default=0),
        ),
        migrations.RunPython(mark_completed_projects)  # manually-reviewed
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-08-04 02:00
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0065_auto_20160802_0301'),
    ]

    operations = [
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='taskassignment',
            unique_together=set([]),
        ),
    ]
